/*
 * SDISKMON 
 * SCSI Disk Monitor utility for Win32 SCSI miniport interface.
 *
 * Author: Andy Cress
 * 
 * 02/04/04 ver 1.0 created with extracts from SDK samples, etc.
 */
typedef unsigned long ULONG_PTR;
#include <windows.h>
#include <winioctl.h>
#include <ntddscsi.h>
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "sdiskmon.h"

#define uchar  unsigned char
#define MAX_DEVS      64
#define DEF_NSEC      5

char progver[] = "1.0";  /* program version */
char fauto   = 0;
char fbackgd   = 0;
char fdebug    = 0;
char finitpass = 1;  /* first time */
long rtime = 5000;   /* 5000 msec = 5 sec */
typedef struct _ThreadBlockInfo {
        HANDLE  hThread;
        LONG    lThreadId;
        short   state;
} THREADBLOCKINFO, *PTHREADBLOCKINFO;
SCSI_PASS_THROUGH_WITH_BUFFERS        sptwb;
THREADBLOCKINFO ThreadParams;
UCHAR seagate_buffer[256];  /* max used = 240 */
FILE *fdlog;
char logfile[] = "sdiskmon.log";
char flogopen = 0;
char filesok  = 1;
ULONG onemeg = 1024*1024;

/* Device Statuses for devlist.dstatus: */
#define  DSTAT_DEAD   0  /*not inserted, dead, gone */
#define  DSTAT_FAIL   1  /*may be inserted, marked failed */
#define  DSTAT_NOTRDY 2  /*inserted, not yet ready (sense key=02)*/
#define  DSTAT_READY  3  /*ready, good sense */
#define  DSTAT_ONL    4  /*online, included in the raid */
char dstat_str[5][8] = {
    "dead  ",  "failed", "notrdy", "ready ", "online"
};
struct {
    uchar chn;
    uchar bus;
    uchar tgt;          /* SCSI target ID */
    uchar lun;
    uchar devtype;      /* device type: 0=disk, 3=proc */
    uchar dstatus;      /* device status, see DSTAT_ above  */
    uchar fses;     /* =1 if SES enabled device */
    uchar fname[64];    /* store fname, device filename */
    uchar vendor[9];
    uchar model[17];
    uchar sernum[13];
    int ret;      /* return code from inquiry scan functions */
    int raidsts;    /* 1=in the raid */
} devlist[MAX_DEVS];

/* Subroutine Prototypes */
VOID PrintStatusResults(BOOLEAN,DWORD,PSCSI_PASS_THROUGH_WITH_BUFFERS,HANDLE);
VOID PrintErrorStatus(PSCSI_PASS_THROUGH_WITH_BUFFERS psptwb );
int  DoInquiry(HANDLE, char *, int);
int  Inquiry2(HANDLE, char *, int);

int logopen()
{
  if (!filesok) {
       fdlog = stdout;
       }
  else if ((fdlog = fopen(logfile,"a+")) == NULL)  {
       printf("Cannot open %s\n",logfile);
       fdlog = stdout;
       }
  else {
       flogopen = 1;
    }
  return(!flogopen);   /* 0 = success */
}

void logclose()
{
  if (flogopen) fclose(fdlog);
}

VOID
doexit(int err)
{
  logclose();
  exit(err);
}

int 
StartUnit(HANDLE fileHandle)
{
 ULONG  length, returned;
 BOOLEAN        status;
 UCHAR *dbuf;
 int dataLength;
 PSCSI_PASS_THROUGH_WITH_BUFFERS psptwb;
 PSENSE_DATA senseBuffer;
 char estr[80];

 psptwb = &sptwb;
 ZeroMemory(psptwb,sizeof(SCSI_PASS_THROUGH_WITH_BUFFERS));

 dataLength = 0;
 psptwb->spt.Length = sizeof(SCSI_PASS_THROUGH);
 psptwb->spt.PathId = 0;
 psptwb->spt.TargetId = 0;      /* kernel fixes this */
 psptwb->spt.Lun = 0;
 psptwb->spt.CdbLength = CDB6GENERIC_LENGTH;
 psptwb->spt.SenseInfoLength = 26;
 psptwb->spt.DataIn = SCSI_IOCTL_DATA_IN;
 psptwb->spt.DataTransferLength = dataLength;
 psptwb->spt.TimeOutValue = 2;  /* timeout in sec */
 psptwb->spt.DataBufferOffset =
                offsetof(SCSI_PASS_THROUGH_WITH_BUFFERS,ucDataBuf);
 psptwb->spt.SenseInfoOffset =
                offsetof(SCSI_PASS_THROUGH_WITH_BUFFERS,ucSenseBuf);
 psptwb->spt.Cdb[0] = 0x1B;  /* SCSI Start Unit cmd */
 psptwb->spt.Cdb[1] = 0x0;   /* lun << 5, usu lun=0 */
 psptwb->spt.Cdb[2] = 0x00;
 psptwb->spt.Cdb[3] = 0x00;
 psptwb->spt.Cdb[4] = 0x01;
 psptwb->spt.Cdb[5] = 0x00;  
 length = offsetof(SCSI_PASS_THROUGH_WITH_BUFFERS,ucDataBuf) +
                psptwb->spt.DataTransferLength;
 status = DeviceIoControl(fileHandle,
                          IOCTL_SCSI_PASS_THROUGH,
                          psptwb,
                          length,
                          psptwb,
                          length,
                          &returned,
                          FALSE);

 if (status && (psptwb->spt.ScsiStatus == 0)) {
    dbuf = (UCHAR *)psptwb->ucDataBuf;
    if (fdebug)  printf("\nStartUnit: %02x %02x %02x %02x\n",
                    dbuf[0],dbuf[1],dbuf[2],dbuf[3]);
    }
 else {
    senseBuffer = (PSENSE_DATA)psptwb->ucSenseBuf;
    sprintf(estr,"StartUnit status %d, sense %x %x %x\n",
        senseBuffer->ErrorCode,
        senseBuffer->SenseKey,
        senseBuffer->AdditionalSenseCode,
        senseBuffer->AdditionalSenseCodeQualifier);
    if (!fbackgd) printf(estr);
    if (flogopen) fprintf(fdlog,estr);
    }
 return(status);
} /*end StartUnit*/

ULONG 
GetDefects(HANDLE fileHandle)
{
 ULONG  length, returned;
 BOOLEAN        status;
 UCHAR *dbuf;
 int dataLength, numGDL;
 UCHAR dfmt;
 // char outstr[80];
 PSCSI_PASS_THROUGH_WITH_BUFFERS psptwb;

 psptwb = &sptwb;
 numGDL = 0xFFFFFFFF;
 /* 
  * Defect format:
  * 0 = block fmt, 4 = index fmt, 5 = phys sector fmt, 6 = vendor fmt 
  */
 dfmt = 0x05;  
 ZeroMemory(psptwb,sizeof(SCSI_PASS_THROUGH_WITH_BUFFERS));

 dataLength = 4;
 psptwb->spt.Length = sizeof(SCSI_PASS_THROUGH);
 psptwb->spt.PathId = 0;
 psptwb->spt.TargetId = 0;      /* kernel fixes this */
 psptwb->spt.Lun = 0;
 psptwb->spt.CdbLength = CDB10GENERIC_LENGTH;
 psptwb->spt.SenseInfoLength = 26;
 psptwb->spt.DataIn = SCSI_IOCTL_DATA_IN;
 psptwb->spt.DataTransferLength = dataLength;
 psptwb->spt.TimeOutValue = 2;  /* timeout in sec */
 psptwb->spt.DataBufferOffset =
                offsetof(SCSI_PASS_THROUGH_WITH_BUFFERS,ucDataBuf);
 psptwb->spt.SenseInfoOffset =
                offsetof(SCSI_PASS_THROUGH_WITH_BUFFERS,ucSenseBuf);
 psptwb->spt.Cdb[0] = 0x37;  /* SCSI Read Defect List cmd */
 psptwb->spt.Cdb[1] = 0x0;   /* lun << 5, usu lun=0 */
 psptwb->spt.Cdb[2] = 0x08 | dfmt;   /* 0x10 for PList, 0x08 for Glist */ 
 psptwb->spt.Cdb[3] = 0x0;
 psptwb->spt.Cdb[4] = 0x0;
 psptwb->spt.Cdb[5] = 0x0;  
 psptwb->spt.Cdb[6] = 0x0;  
 psptwb->spt.Cdb[7] = 0x0;    /*Allocation len*/
 psptwb->spt.Cdb[8] = dataLength;   /*len */
 psptwb->spt.Cdb[9] = 0x0;
 /* 0x25 = Read Capacity, reqlen-0, reply len = 8 */
 /* 0x25, 0, 0, 0, 0, 0, 0, 0, 0, 0 */
 length = offsetof(SCSI_PASS_THROUGH_WITH_BUFFERS,ucDataBuf) +
                psptwb->spt.DataTransferLength;
 status = DeviceIoControl(fileHandle,
                          IOCTL_SCSI_PASS_THROUGH,
                          psptwb,
                          length,
                          psptwb,
                          length,
                          &returned,
                          FALSE);

 if (status && (psptwb->spt.ScsiStatus == 0)) {
    dbuf = (UCHAR *)psptwb->ucDataBuf;
    if (fdebug) {
      printf("GDL: %02x %02x %02x %02x\n",dbuf[0],dbuf[1],dbuf[2],dbuf[3]);
      }
    numGDL = ((dbuf[2]<<8) | dbuf[3]) / 8;
    }
 else {
    PrintStatusResults(status,returned,psptwb,fileHandle);
    // if (!status) PrintErrorStatus(psptwb);
    }
 return(numGDL);
}  /* end GetDefects()*/

VOID 
SetFaultLed( int idisk, int fon)
{
    char outstr[80];
    char astr[50];
    char aparm;

    /* set the disk fault led on Chesnee */
    if (idisk == 1) aparm = 'b';
    else aparm = 'a';
    if (fbackgd) sprintf(astr,">>%s 2>&1",logfile);
    else astr[0] = 0;  /*null string*/
    sprintf(outstr,"alarms -%c%d %s",aparm,fon,astr);
    if (flogopen) fprintf(fdlog,"%s\n",outstr);
    logclose();
    system(outstr);
    logopen();
}

VOID __cdecl
main(int argc, char *argv[])
{
    DWORD           accessMode,
                    shareMode;
    HANDLE          fileHandle;
    UCHAR           string[40];
    ULONG           errorCode;
    BOOLEAN         changeSet = FALSE;
    ULONG           i, maxDisk;
    ULONG           driveNumber;
    int       prevTID;
    int       path, j; 
    char      outstr[80];
    char      inqstr[48];
    char      servorev[9];
    char      serialno[13];
    char      fwrev[5];
    char    dstr[9];
    char    *pstr;
    int     nsec = DEF_NSEC;
    int     ret, ret2;
    ULONG   ltime1;
    ULONG   numd;
    UCHAR   devtype;
    UCHAR   newstat;
    int fchanged;

  if (argc > 1)
  for (j = 1; j < argc; j++)
  {
    if (argv[j][0] == '-' || argv[j][0] == '/') 
    switch (argv[j][1]) 
    {
      case 'a':
        fauto = 1;     /* autmatically repeat/loop (foreground) */
        break;
      case 'b':
        fbackgd = 1;     /* run in background and loop */
        fauto = 1;
        fdebug = 0;      /* can't do debug printfs in background */
        break;
      case 'e':
        filesok = 0;   /* don't write to files (i.e. log) */
        break;
      case 't':
        if (++j < argc) {
           nsec = atoi(argv[j]);
           if (nsec == 0) nsec = DEF_NSEC;
           }
        break;
      case 'v':
        fdebug = 1;     /* verbose/debug */
        break;
      case '?': 
      default: 
        printf("Usage: %s [-a -b -e -t -v] \n", argv[0] );
        printf("         -a = automatically repeat\n");
        printf("         -b = run in background in a repeat loop\n");
        printf("         -e = don't write to sdiskmon.log file.\n");
        printf("         -t 5 = time to sleep between loops, default 5 sec.\n");
        printf("         -v = show verbose & debug messages\n");
        doexit(1);
    } /*end switch*/
  }  /*end argc loop*/

  rtime = nsec * 1000;
  shareMode = FILE_SHARE_READ | FILE_SHARE_WRITE;  // default
  accessMode = GENERIC_WRITE | GENERIC_READ;       // default
  logopen();
  if (flogopen) {
       time((time_t *)&ltime1);
       fprintf(fdlog,"\n--- SDISKMON v%s log started at %s",progver,
               ctime((long *)&ltime1));   // ctime outputs a '\n' at end
       }

  printf("\nIntel Windows SCSI Disk Monitor Utility, version %s\n",progver);
  printf("Copyright (c) 2004 Intel Corporation, by Andy Cress\n");

  printf("Note that the physical disks are listed below in the same order \nas in Disk Administrator.\n");
  driveNumber = 0;
  maxDisk = 0;

  sprintf(outstr,"\nDisk# |Path:TID:LUN|  Description             |FW ver| Serial#  | Status | #Def\n");
  printf(outstr); 
  if (flogopen) fprintf(fdlog,outstr);

  while(TRUE)
  {          /* select disk */

    path = 0;  /* init to zero, not changed by spt */
    prevTID = 0;  /* init to zero, not changed by spt */

    for (i=0;;i++) {
        sprintf(string, "\\\\.\\PhysicalDrive%d", i);

        fileHandle = CreateFile(string,
                                accessMode,
                                shareMode,
                                NULL,
                                OPEN_EXISTING,
                                0,
                                NULL);

        if (fileHandle == INVALID_HANDLE_VALUE) {
            maxDisk = i;
            if (i == 0) {
                sprintf(outstr,"Error opening PhysicalDrive%d. Error: %d\n",
                   i,errorCode = GetLastError());
                if (finitpass || !fbackgd) { 
                    printf(outstr); PrintError(errorCode);
                }
                if (flogopen) fprintf(fdlog,outstr);
                return;
            } else {
                break;
            }

        } else {
            // Get Normal Inquiry string
            strcpy(inqstr,  "----");  
            strcpy(servorev,"    ");  /* servo stub */
            ret = DoInquiry(fileHandle,inqstr,sizeof(inqstr));
            // Get Seagate Servo version 
            if ((ret == 0) && strncmp(&inqstr[8],"SEAGATE",7) == 0) 
              {
               ret2 = Inquiry2(fileHandle,string,sizeof(string));
               if (ret2 == 0) 
                 strncpy(servorev,&string[16],8);
               else strcpy(&servorev[4],&inqstr[36]);
              }
            else {
              strcpy(&servorev[4],&inqstr[36]);
               }
            servorev[8] = 0;
            strncpy(&fwrev[0],&inqstr[32],4);
            fwrev[4] = 0;
            if (inqstr[44] >= '0') {
            strncpy(serialno,&inqstr[36],12);
            serialno[12] = 0;
            } else {
            strncpy(serialno,&inqstr[36],8);
            serialno[8] = 0;
            }

            if (fdebug) {   
               printf("Inq: %x %x %x %x %x %x %x %x - %s\n",
                 inqstr[0],inqstr[1],inqstr[2],inqstr[3],
                 inqstr[4],inqstr[5],inqstr[6],inqstr[7],
                 &inqstr[8]);   
              }  

            devtype = inqstr[0] & 0x1f;
            inqstr[32] = 0;  /* max len for Inquiry string */

            /* Check another function, like the defects or capacity */
            strcpy(dstr," ");
            if (ret == 0) {
              newstat = DSTAT_READY; 
              if (devtype == 0) {  /* if it is a live disk */
                numd = GetDefects(fileHandle);
                if (numd != 0xffffffff) { /*not error*/
                    sprintf(dstr,"%ld",numd);
                    newstat = DSTAT_READY; 
                } else {
                    newstat = DSTAT_FAIL; 
                }
              }
            } else 
              newstat = DSTAT_DEAD; 

            /* check for changes */
            fchanged = 0;
            if (!finitpass && devlist[i].dstatus != newstat) {
              fchanged = 1;
              pstr = " ";
              if ((newstat < DSTAT_READY) && 
                  (devlist[i].dstatus >= DSTAT_READY)) {
                 pstr = "removed";
                 SetFaultLed(i,1);
                 }
              else if ((newstat >= DSTAT_READY) && 
                       (devlist[i].dstatus < DSTAT_READY)) {
                 pstr = "inserted";
                 SetFaultLed(i,0);
                 }
              sprintf(outstr,"Disk%d changed %d-%d: %s\n",
                    i,devlist[i].dstatus,newstat,pstr);
              if (!fbackgd) printf(outstr);
              if (flogopen) fprintf(fdlog,outstr);
            }

            /* Check Target ID for path info */
            /* path = sptwb.spt.PathId; /* old, PathId never changes */
            if (prevTID > sptwb.spt.TargetId)  path++;  
            if (finitpass || fchanged || fdebug) {
             pstr = dstat_str[newstat];
             /* Display disk entry */
             sprintf(outstr,"Disk%d | %d : %2d : %d | %s | %s | %s | %s | %s\n",
               i, path, sptwb.spt.TargetId, sptwb.spt.Lun, 
               &inqstr[8],fwrev, serialno,pstr,dstr);
             if (finitpass || !fbackgd) printf(outstr);
             if (flogopen) fprintf(fdlog,outstr);
            }

            /* save this info in a table */
            devlist[i].bus = 0;
            devlist[i].chn = path;
            devlist[i].tgt = sptwb.spt.TargetId; 
            devlist[i].lun = sptwb.spt.Lun; 
            devlist[i].dstatus = newstat;
            /* copy inqstr, serialno to devlist, if valid */
            if (newstat > DSTAT_DEAD) {
              devlist[i].devtype = devtype;
              strcpy(devlist[i].sernum,serialno);
              strncpy(devlist[i].vendor,&inqstr[8],8);
              devlist[i].vendor[8] = 0;
              strncpy(devlist[i].model,&inqstr[16],16);
              devlist[i].model[16] = 0;
            }

            CloseHandle(fileHandle);
#ifdef DEBUG
            printf("Inquiry2: %s \n", &string[4]); 
#endif
            prevTID = sptwb.spt.TargetId;
        }
    }  /* for each valid disk[i] */

    if (finitpass) {
        sprintf(outstr,"monitoring %d disks ...\n",i);
        printf(outstr);
        if (flogopen) fprintf(fdlog,outstr);
        /* if (fbackgd) { } /* close stdout */
    }
    finitpass = 0;  /* no longer first time */
    if (fauto == 1) {
        logclose();
        Sleep(rtime);
        logopen();
    } else doexit(0);
  }  /* end while fauto */
  doexit(0);

} /*end main()*/


int 
DoInquiry(HANDLE fileHandle, char *pout, int sout)
{
 long  length, returned;
 BOOLEAN        status;
 PSCSI_PASS_THROUGH_WITH_BUFFERS psptwb;

 psptwb = &sptwb;
 ZeroMemory(psptwb,sizeof(SCSI_PASS_THROUGH_WITH_BUFFERS));
 // if (sout > 40) sout = 40;
 if (sout > 80) sout = 80; /* ucDataBuf is 512 bytes */

 psptwb->spt.Length = sizeof(SCSI_PASS_THROUGH);
 psptwb->spt.PathId = 0;
 psptwb->spt.TargetId = 0;      /* kernel fixes this */
 psptwb->spt.Lun = 0;
 psptwb->spt.CdbLength = CDB6GENERIC_LENGTH;
 psptwb->spt.SenseInfoLength = 24;
 psptwb->spt.DataIn = SCSI_IOCTL_DATA_IN;
 psptwb->spt.DataTransferLength = sout;
 psptwb->spt.TimeOutValue = 2;
 psptwb->spt.DataBufferOffset =
                offsetof(SCSI_PASS_THROUGH_WITH_BUFFERS,ucDataBuf);
 psptwb->spt.SenseInfoOffset =
                offsetof(SCSI_PASS_THROUGH_WITH_BUFFERS,ucSenseBuf);
 psptwb->spt.Cdb[0] = SCSIOP_INQUIRY;
 psptwb->spt.Cdb[1] = 0x0;
 psptwb->spt.Cdb[2] = 0x0;
 psptwb->spt.Cdb[3] = 0x0;
 psptwb->spt.Cdb[4] = sout;
 psptwb->spt.Cdb[5] = 0x0;
 length = offsetof(SCSI_PASS_THROUGH_WITH_BUFFERS,ucDataBuf) +
                psptwb->spt.DataTransferLength;

 status = DeviceIoControl(fileHandle,
                          IOCTL_SCSI_PASS_THROUGH,
                          psptwb,
                          sizeof(SCSI_PASS_THROUGH),
                          psptwb,
                          length,
                          &returned,
                          FALSE);

 /*copy output from CDB to inquiry buffer*/
 if (sout > returned) sout = returned;
 memcpy(pout,psptwb->ucDataBuf,sout);
 pout[sout] = 0;

 PrintStatusResults(status,returned,psptwb,fileHandle);
 if (!status) {
    return(GetLastError()); /*error*/
    }
 else return(0); /*success*/
}  /* end DoInquiry() */

int
Inquiry2(HANDLE fileHandle,char *pout, int sout)
{
 long  length,
       returned;
 BOOLEAN  status;
 PSCSI_PASS_THROUGH_WITH_BUFFERS psptwb;

 psptwb = &sptwb;

 ZeroMemory(psptwb,sizeof(SCSI_PASS_THROUGH_WITH_BUFFERS));
 psptwb->spt.Length = sizeof(SCSI_PASS_THROUGH);
 psptwb->spt.PathId = 0;
 psptwb->spt.TargetId = 0;      /* kernel fixes this */
 psptwb->spt.Lun = 0;
 psptwb->spt.CdbLength = CDB6GENERIC_LENGTH;
 psptwb->spt.SenseInfoLength = 24;
 psptwb->spt.DataIn = SCSI_IOCTL_DATA_IN;
 psptwb->spt.DataTransferLength = sout;
 psptwb->spt.TimeOutValue = 2;
 psptwb->spt.DataBufferOffset =
                offsetof(SCSI_PASS_THROUGH_WITH_BUFFERS,ucDataBuf);
 psptwb->spt.SenseInfoOffset =
                offsetof(SCSI_PASS_THROUGH_WITH_BUFFERS,ucSenseBuf);
 psptwb->spt.Cdb[0] = 0x12; // special SCSI Inquiry command
 psptwb->spt.Cdb[1] = 0x01; // LUN bits 7:5,EVPD(Enable Vital Product Data) bit0
 psptwb->spt.Cdb[2] = 0xC0; // Page Code - nonzero only if EVPD is set
 psptwb->spt.Cdb[3] = 0x0;  // Reserved
 psptwb->spt.Cdb[4] = sout; // allocation length
 psptwb->spt.Cdb[5] = 0x0;  // no special flags or links
   /* This command returns 16 bytes:  */
   /*   0-7  = SCSI version           */
   /*   8-15 = TMS(servo) version     */
 length = offsetof(SCSI_PASS_THROUGH_WITH_BUFFERS,ucDataBuf) +
                psptwb->spt.DataTransferLength;

 status = DeviceIoControl(fileHandle,
                          IOCTL_SCSI_PASS_THROUGH,
                          psptwb,
                          sizeof(SCSI_PASS_THROUGH),
                          psptwb,
                          length,
                          &returned,
                          FALSE);

 /*copy output to inquiry buffer*/
 if (sout > returned) sout = returned;
 memcpy(pout,psptwb->ucDataBuf,sout);
 pout[sout] = 0;

 PrintStatusResults(status,returned,psptwb,fileHandle);
 if (!status) {
    return(GetLastError()); /*error*/
    }
 else return(0); /*success*/
}  /* end Inquiry2() */


VOID
PrintError(ULONG ErrorCode)
{
    UCHAR errorBuffer[80];
    ULONG count;

    if (fbackgd) return;
    count = FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM,
                  NULL,
                  ErrorCode,
                  0,
                  errorBuffer,
                  sizeof(errorBuffer),
                  NULL
                  );
    if (count != 0) {
        printf("%s\n", errorBuffer);
    } else {
        printf("FormatMessage failed.  Error: %d\n", GetLastError());
    }
}   /* end PrintError() */

VOID
PrintSenseInfo(PSCSI_PASS_THROUGH_WITH_BUFFERS psptwb)
{
    PSENSE_DATA senseBuffer = (PSENSE_DATA)psptwb->ucSenseBuf;
    int e,k,a,q;
    UCHAR *dbuf;

    if (fdebug && !fbackgd) printf("SCSI request sense data\n");

    e = senseBuffer->ErrorCode;
    k = senseBuffer->SenseKey;
    a = senseBuffer->AdditionalSenseCode;
    q = senseBuffer->AdditionalSenseCodeQualifier;
    if (flogopen) 
      fprintf(fdlog,"Error %x, SCSI Sense data: %x %x %x\n",e,k,a,q);

    if (fbackgd) return;
    if (psptwb->spt.SenseInfoLength == 0) {
       printf("No sense data\n");
       return;
    }

    printf("Error %x, SCSI Sense data: %x %x %x\n",e,k,a,q);
    k = k & 0x0f;

    switch (k) {

    case SCSI_SENSE_NOT_READY:
        printf("Device not ready, ");
        switch (a) {
        case SCSI_ADSENSE_LUN_NOT_READY:
            printf("Lun not ready, ");
            switch (q) {
            case SCSI_SENSEQ_BECOMING_READY:
                printf("In process of becoming ready\n");
                break;
            case SCSI_SENSEQ_MANUAL_INTERVENTION_REQUIRED:
                printf("Manual intervention required\n");
                break;
            case SCSI_SENSEQ_FORMAT_IN_PROGRESS:
                printf("Format in progress\n");
                break;
            case SCSI_SENSEQ_INIT_COMMAND_REQUIRED:
                printf("Initializing command required\n");
                break;
            default:
                printf("ASCQ = %x\n",q);
                break;
            } // end switch (senseBuffer->AdditionalSenseCodeQualifier)
            break;
        case SCSI_ADSENSE_NO_MEDIA_IN_DEVICE:
            printf(" No Media in device.\n");
            break;
        default:
            printf("ASC = %x\n",a);
        } // end switch (senseBuffer->AdditionalSenseCode)
        break;

    case SCSI_SENSE_DATA_PROTECT:
        printf("Media write protected\n");
        break;

    case SCSI_SENSE_MEDIUM_ERROR:
        printf("Bad media\n");
        break;

    case SCSI_SENSE_HARDWARE_ERROR:
        printf("Hardware error\n");
        break;

    case SCSI_SENSE_ILLEGAL_REQUEST:
        printf("Illegal SCSI request, ");
        dbuf = psptwb->ucSenseBuf;
        if ((dbuf[2] & 0x0f) == 5 && (dbuf[15] & 0x80) == 0x80) { /*SKSV set*/
           unsigned short l;
           (dbuf[15] & 0x40) == 0x40 ? printf("Error in CDB: ")
                          : printf("Error in Parameter List: ");
           l = dbuf[16] * 0x100 + dbuf[17];
           printf("Byte #%d", l);
           (dbuf[15] & 0x08) == 0x08 ?   /* BPV set? (Bit Pointer Valid) */
           printf(", Bit #%d is in error.\n", dbuf[15] & 0x07) :
           printf(" is in error.\n");                            
        } else
        switch (a) {
        case SCSI_ADSENSE_ILLEGAL_COMMAND:
            printf("Illegal command\n");
            break;
        case SCSI_ADSENSE_ILLEGAL_BLOCK:
            printf("Illegal block address\n");
            break;
        case SCSI_ADSENSE_INVALID_LUN:
            printf("Invalid LUN\n");
            break;
        case SCSI_ADSENSE_MUSIC_AREA:
            printf("Music area\n");
            break;
        case SCSI_ADSENSE_DATA_AREA:
            printf("Data area\n");
            break;
        case SCSI_ADSENSE_VOLUME_OVERFLOW:
            printf("Volume overflow\n");
            break;
        case SCSI_ADSENSE_INVALID_CDB:
            printf("Invalid field in CDB\n");
            break;
        } // end switch (senseBuffer->AdditionalSenseCode)
        break;

    case SCSI_SENSE_UNIT_ATTENTION:
        switch (a) {
        case SCSI_ADSENSE_MEDIUM_CHANGED:
            printf("Media changed\n");
            break;
        case SCSI_ADSENSE_BUS_RESET:
            printf("Bus reset\n");
            break;
        default:
            printf("Unit attention\n");
            break;
        } // end  switch (senseBuffer->AdditionalSenseCode)
        break;

    case SCSI_SENSE_ABORTED_COMMAND:
        printf("Command aborted\n");
        break;

    case SCSI_SENSE_RECOVERED_ERROR:
        printf("Recovered error\n");
        switch(a) {
        case SCSI_ADSENSE_SEEK_ERROR:
            printf("Seek error\n");
            break;
        case SCSI_ADSENSE_TRACK_ERROR:
            printf("Track error\n");
            break;
        case SCSI_ADSENSE_REC_DATA_NOECC:
            printf("Recovered data - no ECC\n");
            break;
        case SCSI_ADSENSE_REC_DATA_ECC:
            printf("Recovered data - ECC\n");
            break;
        default:
            break;
        } // end switch(senseBuffer->AdditionalSenseCode)
        if (senseBuffer->IncorrectLength) {
            printf("Incorrect length detected.\n");
        }
        break;

    case SCSI_SENSE_NO_SENSE:
        //
        // Check other indicators.
        //
        if (senseBuffer->IncorrectLength) {
            printf("Incorrect length detected.\n");
        } else {
            printf("No specific sense key\n");
        }
        break;

    default:
        printf("Unrecognized sense code\n");
        break;

    } // end switch (senseBuffer->SenseKey & 0xf)

}  /*end PrintSenseInfo()*/

VOID
PrintStatusResults(
    BOOLEAN status,DWORD returned,PSCSI_PASS_THROUGH_WITH_BUFFERS psptwb,
    HANDLE fh)
{
    ULONG errorCode;
    PSENSE_DATA senseBuffer = (PSENSE_DATA)psptwb->ucSenseBuf;
    char estr[80];

    if (!status) {
       sprintf(estr, "Error: %d  ", errorCode = GetLastError() );
       if (!fbackgd) { 
           printf(estr);
           PrintError(errorCode);
       }
       if (flogopen) fprintf(fdlog,"%s\n",estr);
       return;
    }
    if (psptwb->spt.ScsiStatus) {
       int k,a,q;
       k = senseBuffer->SenseKey;
       a = senseBuffer->AdditionalSenseCode;
       q = senseBuffer->AdditionalSenseCodeQualifier;
       PrintSenseInfo(psptwb);
       if (k == 2 && a == 4 && q == 2) StartUnit(fh);
       return;
    } else {
       /* command successful */
       return;
    }
}

VOID
PrintInquiryStatusResults(
    BOOLEAN status,DWORD returned,PSCSI_PASS_THROUGH_WITH_BUFFERS psptwb,
    ULONG length)
{
    ULONG errorCode;
    char estr[80];

    if (!status ) {
       sprintf(estr, "Error: %d  ", errorCode = GetLastError() );
       if (!fbackgd) { 
           printf(estr);
           PrintError(errorCode);
       }
       if (flogopen) fprintf(fdlog,"%s\n",estr);
       return;
    }
    if (psptwb->spt.ScsiStatus) {
       PrintSenseInfo(psptwb);
       return;
    } else {
       if (!fbackgd)
          printf("%s", &psptwb->ucDataBuf[8]);
       return;
    }
}


VOID
PrintErrorStatus(
    PSCSI_PASS_THROUGH_WITH_BUFFERS psptwb
    )

{
    PrintError(GetLastError());

    if (psptwb->spt.ScsiStatus) {
       PrintSenseInfo(psptwb);
       return;
    }
}

/* end sdiskmon.c */
